/*
 * Copyright (c) 2009,2010 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 ** rmp400MotionTaskCodels.c
 **
 ** Codels called by execution task rmp400MotionTask
 **
 ** Author: Matthieu Herrb
 ** Date: April 2009
 **
 **/
#include <sys/types.h>
#include <errno.h>
#include <math.h>
#include <pthread.h>
#include <stdlib.h>

#include <portLib.h>
#include <h2mathLib.h>

#include <rmp/rmpLib.h>

#include "rmp400Const.h"
#include "server/rmp400Header.h"

#include "odoProba.h"
#include "joystick.h"

static struct RMP_DEV_STR **rmp;
static POSTER_ID pomPosterId = NULL;
static JOYSTICK_STR *joy = NULL;
extern void *rmp400ReadTask(void *);

static STATUS rmp400DataUpdate(RMP400_STATUS_STR *);
static STATUS rmp400VelocitySet(double, double, int *);
static STATUS initPomPosters(void);
static STATUS updatePomPosters(void);
static STATUS rmp400VelocityGet(void);
static void getJoystickSpeeds(double *v, double *w);
/*------------------------------------------------------------------------
 *
 * initOdoAndAsserv  --  Initialization codel (fIDS, ...)
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
initOdoAndAsserv(int *report)
{
	GENPOS_CART_STATE *configId;
	GENPOS_CART_REF *refId;
	GENPOS_CART_CMD_PARAM *cmdId;
	RMP400_KINEMATICS_STR *kinematics;

	/* Kinematics */
	kinematics = &rmp400DataStrId->kinematics;
	kinematics->leftWheelRadius = RMP400_LEFT_WHEEL_RADIUS;
	kinematics->rightWheelRadius = RMP400_RIGHT_WHEEL_RADIUS;
	kinematics->axisWidth = RMP400_AXIS_WIDTH;

	/* configuration */
	configId = &rmp400DataStrId->robot;
	configId->xRob = 0.;
	configId->yRob = 0.;
	configId->theta = 0.;
	configId->xRef = RMP400_DIST_TO_CONTROLLED_POINT;
	configId->yRef = 0;
	configId->v = 0;
	configId->w = 0;

	/* Ref Data */
	refId = &rmp400DataStrId->ref;
	refId->dataType = GENPOS_POS_AND_SPEED_DATA;
	refId->backFlag = GENPOS_FORWARD_MOTION;
	refId->x = configId->xRef;
	refId->y = configId->yRef;
	refId->theta = configId->theta;
	refId->v = 0.;
	refId->w = 0.;

	refId->vmax = RMP400_VMAX;
	refId->wmax = RMP400_WMAX;
	refId->linAccelMax = RMP400_ACCEL_LIN_MAX;
	refId->angAccelMax = RMP400_ACCEL_ANG_MAX;

	/* Control parameters */
	cmdId = &rmp400DataStrId->cmd;
	cmdId->distPoint = RMP400_DIST_TO_CONTROLLED_POINT;
	cmdId->KpLongit = RMP400_KP_LONGIT;
	cmdId->KiLongit = RMP400_KI_LONGIT;
	cmdId->KpTransv = RMP400_KP_TRANSV;
	cmdId->KiTransv = RMP400_KI_TRANSV;

	cmdId->KpAng = RMP400_KP_ANG;
	cmdId->KiAng = RMP400_KI_ANG;

        cmdId->longitMaxGap = RMP400_LONGIT_MAX_GAP;
        cmdId->transvMaxGap = RMP400_TRANSV_MAX_GAP;
        cmdId->angMaxGap    = RMP400_ANG_MAX_GAP;

	cmdId->longitErrorMax = RMP400_LONGIT_ERROR_MAX;
	cmdId->transvErrorMax = RMP400_TRANSV_ERROR_MAX;
	cmdId->angErrorMax = RMP400_ANG_ERROR_MAX;

	cmdId->longitMaxErrSum = RMP400_LONGIT_MAX_ERR_SUM;
	cmdId->transvMaxErrSum = RMP400_TRANSV_MAX_ERR_SUM;
	cmdId->angMaxErrSum = RMP400_ANG_MAX_ERR_SUM;

	/* Initialize Pom */
	initPomPosters();

	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * rmp400DataUpdate - copy data from 2 motors block to the IDS
 */
static STATUS
rmp400DataUpdate(RMP400_STATUS_STR *status)
{
	static int prevFrame[2] = {0, 0};
	struct RMP_DATA_STR *data;
	struct RMP400_DATA_STR *dest;
	int i;

	for (i = 0; i < 2; i++) {
		data = rmpGetData(rmp[i]);
		if (data == NULL) {
			fprintf(stderr, "NULL state str %d\n", i);
			continue;
		}
#if 0
		if (data->servo_frames == prevFrame[i]) {
			fprintf(stderr, 
				"Data not refreshed by rmp %d: %d\n", 
				i, data->servo_frames);
			return ERROR;
		}
#endif
		prevFrame[i] = data->servo_frames;
		dest = &status->rs_data[i];

		pthread_mutex_lock(rmpGetMutex(rmp[i]));

		dest->pitch_angle = RMP400_CONVERT_PITCH(data->pitch_angle);
		dest->pitch_rate = RMP400_CONVERT_PITCH(data->pitch_rate);
		dest->roll_angle = RMP400_CONVERT_ROLL(data->roll_angle);
		dest->lw_velocity = RMP400_CONVERT_WHEEL_SPEED(data->lw_velocity);
		dest->rw_velocity = RMP400_CONVERT_WHEEL_SPEED(data->rw_velocity);
		dest->yaw_rate = RMP400_CONVERT_YAW(data->yaw_rate);
		dest->servo_frames = data->servo_frames;
		dest->integrated_left_wheel = 
		   RMP400_CONVERT_INTEGRATED_POS(data->integrated_left_wheel);
		dest->integrated_right_wheel = 
		   RMP400_CONVERT_INTEGRATED_POS(data->integrated_right_wheel);
		dest->integrated_fore_aft = 
		   RMP400_CONVERT_INTEGRATED_POS(data->integrated_fore_aft);
		dest->integrated_yaw = 
		   RMP400_CONVERT_INTEGRATED_TURN(data->integrated_yaw);
		dest->left_torque = RMP400_CONVERT_TORQUE(data->left_torque);
		dest->right_torque = RMP400_CONVERT_TORQUE(data->right_torque);
		dest->operational_mode = data->operational_mode;
		dest->controller_gain_schedule =
		    data->controller_gain_schedule;
		dest->ui_voltage = RMP400_CONVERT_UI_VOLTAGE(data->ui_voltage);
		dest->powerbase_voltage = 
                    RMP400_CONVERT_POWERBASE_VOLTAGE(data->powerbase_voltage);
		dest->velocity_command = data->velocity_command; /* ?? */
		dest->turn_command = data->turn_command; /* ?? */

		pthread_mutex_unlock(rmpGetMutex(rmp[i]));
	}
	return OK;
}



/*------------------------------------------------------------------------
 *
 * odo  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description: Reads all robots state and computes odometry
 *
 * Reports:      OK
 *              S_rmp400_EMERGENCY_STOP
 *
 * Returns:    OK or ERROR
 */

STATUS
odo(int *report)
{
	GENPOS_CART_REF *refId = &SDI_F->ref;
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	RMP400_STATUS_STR *status = &SDI_F->status;
	double direction;

	if (rmp == NULL || rmp[0] == NULL || rmp[1] == NULL)
		return OK;		/* not initialized yet */

	rmp400DataUpdate(status);

	/* Check motors status */
	if (status->rs_mode == RMP400_MOTORS_OFF && 
	    status->rs_data[0].operational_mode == 1 &&
	    status->rs_data[1].operational_mode == 1) {
		printf("Motors ON\n");
		status->rs_mode = RMP400_IDLE;
	}
	/* XXX Need to Check if motors went OFF */

	/* Read config */
	if (rmp400VelocityGet() != OK)
		return ERROR;
	if (refId->backFlag == GENPOS_FORWARD_MOTION)
		direction = 1;
	else if (refId->backFlag == GENPOS_BACKWARD_MOTION)
		direction = -1;
	else
		direction = 0;
	robot->xRef = robot->xRob 
		+ direction*RMP400_DIST_TO_CONTROLLED_POINT*cos(robot->theta);
	robot->yRef = robot->yRob 
		+ direction*RMP400_DIST_TO_CONTROLLED_POINT*sin(robot->theta);

	odoProba(&SDI_F->robot, &SDI_F->var,
	    SDI_F->kinematics.axisWidth, SDI_F->varParams.coeffLinAng,
		   EXEC_TASK_PERIOD(RMP400_MOTIONTASK_NUM));
	return OK;
}

/*------------------------------------------------------------------------
 *
 * asserv  --  2nd codel of permanent activity (called AFTER other activities)
 *
 * Description: Computes reference speeds and apply them
 * 
 * Reports:      OK
 *              S_rmp400_EMERGENCY_STOP
 * 
 * Returns:    OK or ERROR
 */

STATUS
asserv(int *report)
{
	GENPOS_CART_REF *refId = &SDI_F->ref;
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	double vRef, wRef,dLongit, dTransv;
	static double prevV, prevW;
	static BOOL prevDrifted = FALSE;
	BOOL drifted = FALSE;

	if (rmp == NULL || rmp[0] == NULL || rmp[1] == NULL)
		return OK;		/* not initialized yet */

	switch (SDI_F->status.rs_mode) {
	case RMP400_MOTORS_OFF:
	case RMP400_EMERGENCY:
		/* No motion possible */
		break;
		
	case RMP400_IDLE:
		vRef = 0.0;
		wRef = 0.0;
		refId->x = robot->xRef;
		refId->y = robot->yRef;
		refId->theta = robot->theta;
		break;

	case RMP400_MANUAL:
		if (joystickPoll(joy) < 0) {
			*report = S_rmp400_JOYSTICK_ERROR;
			SDI_F->status.rs_mode = RMP400_IDLE;
			break;
		}
		refId->x = robot->xRef;
		refId->y = robot->yRef;
		refId->theta = robot->theta;
		getJoystickSpeeds(&vRef, &wRef);
		refId->v = vRef;
		refId->w = wRef;
		break;

	case RMP400_TRACK_POSITION:
		if (refId->dataType != GENPOS_POS_DATA) {
			fprintf(stderr, "wrong data type for TRACK_POS\n");
			errnoSet(S_rmp400_BAD_REF);
			break;
		}
		drifted =  contLawCartPositionControl(&rmp400DataStrId->cmd,
			refId, robot, EXEC_TASK_PERIOD(RMP400_MOTIONTASK_NUM),
			prevV, prevW, &vRef, &wRef);
		break;

	case RMP400_TRACK_CONFIG:
		if (refId->dataType != GENPOS_POS_AND_SPEED_DATA) {
			fprintf(stderr, "wrong data type for TRACK_CONFIG\n");
			errnoSet(S_rmp400_BAD_REF);
			break;
		}
		drifted = contLawCartConfigControl(&rmp400DataStrId->cmd,
		    refId, robot, EXEC_TASK_PERIOD(RMP400_MOTIONTASK_NUM),
		    prevV, prevW, &vRef, &wRef);
		break;

	case RMP400_TRACK_SPEED:
		if (refId->dataType != GENPOS_SPEED_DATA) {
			fprintf(stderr, "wrong data type for TRACK_SPEED %d\n",
				refId->dataType);
			errnoSet(S_rmp400_BAD_REF);
			break;
		}
		vRef = (refId->backFlag == GENPOS_FORWARD_MOTION ? refId->v :
		    -refId->v);
		wRef = refId->w;
		break;

	default:
		errnoSet(S_rmp400_BAD_REF);
		return ERROR;
	}
	prevV = vRef;
	prevW = wRef;
	if (drifted && !prevDrifted) {
		printf("odoAndAsserv: important drift!\n");
		printf("ref: %.2lf %.2lf %.2lf\n",
		       refId->x, refId->y, refId->theta);
		printf("pos: %.2lf %.2lf %.2lf\n",
		       robot->xRef, robot->yRef, robot->theta);
	}
	if (!drifted && prevDrifted)
		printf("odoAndAsserv: drift cancelled\n");
	prevDrifted = drifted;

	if (rmp400VelocitySet(vRef, wRef, report) != OK)
		return ERROR;

	updatePomPosters();

	return OK;
}

/*------------------------------------------------------------------------
 * Init
 *
 * Description:
 *
 * Reports:      OK
 */

/* rmp400InitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400InitStart(int *report)
{
	int i, n;
	char name[20];

	*report = OK;

	if ((n = rmpOpenAll(&rmp)) == -1) {
		*report = S_rmp400_RMPLIB_ERROR;
		return ETHER;
	}
	if (n != 2) {
		/* we expect 2 RMP devices */
		*report = S_rmp400_RMPLIB_ERROR;
		return ETHER;
	}
	printf("rmpOpenAll ok\n");
	/* start read tasks */
	errnoSet(0);
	for (i = 0; i < 2; i++) {
		snprintf(name, sizeof(name), "tReadTask%d", i);
		if (taskSpawn(name, 100, 0, 8192, rmp400ReadTask, rmp[i]) 
		    == ERROR) {
			fprintf(stderr, "taskSpawn %s error\n", name);
			*report =  errnoGet();
			break;
		}
		printf("%s spawned\n", name);
	}
	if (*report != OK) {
		fprintf(stderr, "rmp400InitMain: report 0x%x\n", *report);
		rmpClose(rmp[0]);
		rmpClose(rmp[1]);
		free(rmp);
	}
	joy = joystickInit(RMP400_JOYSTICK_DEVICE_PATH);
	if (joy == NULL) {
		fprintf(stderr, "joystickInit(%s) failed: %s\n",
			RMP400_JOYSTICK_DEVICE_PATH, strerror(errno));
	}
	for (i = 0; i < 2; i++) 
		if (rmpPowerOn(rmp[i]) == ERROR) {
			*report = errnoGet();
			break;
		}
	return EXEC;
}

ACTIVITY_EVENT
rmp400InitMain(int *report)
{
	RMP400_STATUS_STR *status = &SDI_F->status;

	/* Check motors status */
	if (status->rs_data[0].operational_mode != 1 ||
	    status->rs_data[1].operational_mode != 1) {
		printf("Motors OFF\n");
		status->rs_mode = RMP400_MOTORS_OFF;
	} else {
		printf("Motors ON\n");
		status->rs_mode = RMP400_IDLE;
	}
	return ETHER;
}
	
/*------------------------------------------------------------------------
 * JoystickOn
 *
 * Description: 
 *
 * Reports:      OK
 *              S_rmp400_BAD_REF
 *              S_rmp400_RMP_LIB_ERROR
 */

/* rmp400JoystickOnStart  -  codel START of JoystickOn
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400JoystickOnStart(int *report)
{
	if (joy == NULL) {
		*report = S_rmp400_JOYSTICK_ERROR;
		return ETHER;
	}
	SDI_F->status.rs_mode = RMP400_MANUAL;
	return EXEC;
}

/* rmp400JoystickOnMain  -  codel EXEC of JoystickOn
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400JoystickOnMain(int *report)
{
	/* Check for user abort (button 4) */
	if (joy->buttons[3]) {
		printf("Stop joystick\n");
		SDI_F->status.rs_mode = RMP400_IDLE;
		return ETHER;
	}
	return EXEC;
}

/* rmp400JoystickOnInter  -  codel INTER of JoystickOn
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400JoystickOnInter(int *report)
{
	SDI_F->status.rs_mode = RMP400_IDLE;
	return ETHER;
}

/*------------------------------------------------------------------------
 * SetPom
 *
 * Description:
 *
 * Reports:      OK
 *              S_rmp400_POSTER_NOT_FOUND
 */

/* setPom  -  codel EXEC of SetPom
   Returns:  EXEC END ETHER FAIL ZOMBIE */

ACTIVITY_EVENT
setPom(GENPOS_POSTER_NAME *posterName, int *report)
{
	POSTER_ID id;

	if (posterFind(posterName->name, &id) != OK) {
		logMsg("posterFind(%s) failed\n", posterName->name);
		*report = S_rmp400_POSTER_NOT_FOUND;
	} else
		pomPosterId = id;
	return ETHER;
}

/*----------------------------------------------------------------------*/
/*
 * Compute speed
 *
 * sets robot.v and robot.w in the SDI from motion in the motor structs.
 */
static STATUS
rmp400VelocityGet(void)
{
	double leftv, rightv;
	RMP400_DATA_STR *rmpData[2];
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	RMP400_KINEMATICS_STR *kinematics = &SDI_F->kinematics;
	int i;

	for (i = 0; i < 2; i++)
		rmpData[i] = &SDI_F->status.rs_data[i];

	/* Mean angular speed of front and back wheels */
	leftv = (rmpData[0]->lw_velocity + rmpData[1]->lw_velocity) / 2.0;
	rightv = (rmpData[0]->rw_velocity + rmpData[1]->rw_velocity) / 2.0;
	/* printf("%.2lf %.2lf\n", leftv, rightv); */

	/* Angular speed of the robot */
#ifdef USE_INTERNAL_GYRO
	robot->w = (rmpData[0]->yaw_rate + rmpData[1]->yaw_rate) / 2.0;
#else
	robot->w = (leftv - rightv) / kinematics->axisWidth;
#endif
	/* linear speeds */
	robot->v = (leftv + rightv) / 2.0;

	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Send given linear and angular speeds to the hardware
 *
 */
static STATUS
rmp400VelocitySet(double v, double w, int *report)
{
	int i;

	/* Send the same speed to front and back axis */
	for (i = 0; i < 2; i++) {
		if (rmpSendSpeed(rmp[i], RMP400_CONVERT_TO_RMP_SPEED(v),
				 RMP400_CONVERT_TO_RMP_ANGLE(w)) != OK) {
			*report = errnoGet();
			return ERROR;
		}
	}
	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Initialize POM
 */
static STATUS
initPomPosters(void)
{
	POM_ME_POS *mePos = &rmp400DataStrId->mePos;

	mePos->kind = POM_ME_ABSOLUTE;
	mePos->date1 = -1;
	mePos->confidence = 0.2;	// XXX used?
	mePos->main.var.voo     = RMP400_VAR_TT;
	mePos->main.var.vpp     = RMP400_VAR_TT;
	mePos->main.var.vrr     = RMP400_VAR_TT;
	mePos->main.var.vxx     = RMP400_VAR_XX;
	mePos->main.var.vyy     = RMP400_VAR_YY;
	mePos->main.var.vzz     = RMP400_VAR_XX;

	return OK;
}

/*----------------------------------------------------------------------*/

/*
 * Update POM poster
 */
static STATUS
updatePomPosters(void)
{
	POM_ME_POS *mePos = &SDI_F->mePos;
	GENPOS_CART_STATE *configId;
	POM_POS pomPos;

	if (pomPosterId != NULL) {
		if(rmp400POM_POSPosterRead(pomPosterId, &pomPos) == OK) {
			mePos->date1 = pomPos.date;
		} else {
			logMsg("rmp400POM_POSPosterRead() failed\n");
			pomPosterId = NULL;
			mePos->date1 = -1;
		}
	} else
		mePos->date1 = -1;

	configId = &rmp400DataStrId->robot;
	mePos->main.euler.yaw = configId->theta;
	mePos->main.euler.x   = configId->xRob;
	mePos->main.euler.y   = configId->yRob;

	/* speeds are mostly zero, just update the linear
	   and angular components */
	mePos->VLocal.euler.x   = configId->v;
	mePos->VLocal.euler.yaw = configId->w;

	/* all variances and co-variances remain zero */

	return OK;
}

/*----------------------------------------------------------------------*/

/*
 * Convert joystick values to reference speed
 */
static void 
getJoystickSpeeds(double *v, double *w)
{
	if (joy->buttons[0]) {
		*v = joy->axes[1] * RMP400_JOYSTICK_VGAIN;
		*w = joy->axes[0] * RMP400_JOYSTICK_WGAIN;
	}
	else {
		*v = 0;
		*w = 0;
	}
}
